// Copyright 2018 The Dawn & Tint Authors
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
//    contributors may be used to endorse or promote products derived from
//    this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef SRC_DAWN_NATIVE_VULKAN_SHADERMODULEVK_H_
#define SRC_DAWN_NATIVE_VULKAN_SHADERMODULEVK_H_

#include <memory>
#include <mutex>
#include <optional>
#include <string>
#include <utility>

#include "dawn/common/HashUtils.h"
#include "dawn/common/vulkan_platform.h"
#include "dawn/native/Error.h"
#include "dawn/native/ShaderModule.h"
#include "partition_alloc/pointers/raw_ptr.h"

namespace dawn::native {

struct ProgrammableStage;

namespace vulkan {

struct TransformedShaderModuleCacheKey {
    uintptr_t layoutPtr;
    std::string entryPoint;
    PipelineConstantEntries constants;
    std::optional<uint32_t> maxSubgroupSizeForFullSubgroups;

    bool operator==(const TransformedShaderModuleCacheKey& other) const;
};

struct TransformedShaderModuleCacheKeyHashFunc {
    size_t operator()(const TransformedShaderModuleCacheKey& key) const;
};

class Device;
class PipelineLayout;

class ShaderModule final : public ShaderModuleBase {
  public:
    struct ModuleAndSpirv {
        VkShaderModule module;
        const uint32_t* spirv;
        size_t wordCount;
        std::string remappedEntryPoint;
    };

    static ResultOrError<Ref<ShaderModule>> Create(
        Device* device,
        const UnpackedPtr<ShaderModuleDescriptor>& descriptor,
        ShaderModuleParseResult* parseResult,
        OwnedCompilationMessages* compilationMessages);

    ResultOrError<ModuleAndSpirv> GetHandleAndSpirv(
        SingleShaderStage stage,
        const ProgrammableStage& programmableStage,
        const PipelineLayout* layout,
        bool clampFragDepth,
        bool emitPointSize,
        std::optional<uint32_t> maxSubgroupSizeForFullSubgroups);

  private:
    ShaderModule(Device* device, const UnpackedPtr<ShaderModuleDescriptor>& descriptor);
    ~ShaderModule() override;
    MaybeError Initialize(ShaderModuleParseResult* parseResult,
                          OwnedCompilationMessages* compilationMessages);
    void DestroyImpl() override;

    // New handles created by GetHandleAndSpirv at pipeline creation time.
    class ConcurrentTransformedShaderModuleCache;
    std::unique_ptr<ConcurrentTransformedShaderModuleCache> mTransformedShaderModuleCache;
};

}  // namespace vulkan

}  // namespace dawn::native

#endif  // SRC_DAWN_NATIVE_VULKAN_SHADERMODULEVK_H_
